"""This websocket implementation has been DEPRECATED"""

from .libs import setup_pyaudio, teardown_pyaudio, play_audio, send_async_generator
from .neuphonic_websocket_client import NeuphonicWebsocketClient
